/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/loader/obj_loader.h>
#include <blt/std/loader.h>
#include <blt/std/string.h>
#include <blt/std/logging.h>
#include <cctype>
#include <charconv>

namespace blt::gfx
{
    class char_tokenizer
    {
        private:
            std::string_view string;
            std::size_t current_pos = 0;
        public:
            explicit char_tokenizer(std::string_view view): string(view)
            {}
            
            inline char peek()
            {
                return string[current_pos];
            }
            
            inline char advance()
            {
                return string[current_pos++];
            }
            
            inline bool has_next(size_t offset = 0)
            {
                return current_pos + offset < string.size();
            }
            
            inline std::string_view read_fully()
            {
                return blt::string::trim(string.substr(current_pos));
            }
    };
    
    float get(std::string_view str)
    {
        float x;
        const auto [ptr, ec] = std::from_chars(str.data(), str.data() + str.size(), x);
        // probably not needed.
        if (ec != std::errc())
        {
//            int i;
//            const auto [ptr2, ec2] = std::from_chars(str.data(), str.data() + str.size(), i);
//            if (ec2 == std::errc())
//            {
//                x = static_cast<float>(i);
//            } else
//            {
                BLT_WARN("Unable to parse string '%s' into number!", std::string(str).c_str());
                x = 0;
//            }
        }
        return x;
    }
    
    void obj_loader::parse_extra_line(char_tokenizer& tokenizer)
    {
        char type = tokenizer.advance();
        auto elements = blt::string::split(std::string(tokenizer.read_fully()), " ");
        if (elements.size() < 2)
        {
            BLT_ERROR("Unable to parse line '%s' too few arguments to type '%c'", std::string(tokenizer.read_fully()).c_str(), type);
            return;
        }
        float x = get(elements[0]), y = get(elements[1]);
        if (std::isspace(type))
        {
            if (elements.size() < 3)
            {
                BLT_ERROR("Unable to parse line '%s' too few arguments to type '%c'", std::string(tokenizer.read_fully()).c_str(), type);
                return;
            }
            float z = get(elements[2]);
            vertices.push_back(vertex_t{x, y, z});
            return;
        }
        switch (type)
        {
            case 't':
                uvs.push_back(uv_t{x,y});
                break;
            case 'n':
            {
                if (elements.size() < 3)
                {
                    BLT_ERROR("Unable to parse line '%s' too few arguments to type '%c'", std::string(tokenizer.read_fully()).c_str(), type);
                    return;
                }
                float z = get(elements[2]);
                normals.push_back(vertex_t{x, y, z});
                break;
            }
            case 'p':
                // optional todo
                break;
            default:
                BLT_WARN("Unexpected type '%c'", type);
                break;
        }
    }
    
    std::vector<obj_object_t> quick_load(std::string_view file)
    {
        std::vector<obj_object_t> objects;
        
        
        return objects;
    }
    
    void obj_loader::parseFile(std::string_view file)
    {
        auto lines = blt::fs::getLinesFromFile(std::string(file));
        for (const auto& line : lines)
        {
            char_tokenizer token(line);
            if (!token.has_next() || token.read_fully().empty())
                continue;
            switch (token.advance())
            {
                case '#':
                    continue;
                case 'f':
                {
                    
                    break;
                }
                case 'v':
                {
                    parse_extra_line(token);
                    break;
                }
                case 'o':
                {
                
                }
            }
        }
    }
    
    void obj_loader::parse_face(std::string_view line)
    {
    
    }
}